class BinaryOperator:
    def __init__(self, left, right):
        self.left = left
        self.right = right

    def __str__(self):
        return '%s(%s, %s)' % (self.__class__.__name__, self.left, self.right)

    def toSympy(self, op):
        return '(%s%s%s)' % (self.left.toSympy(), op, self.right.toSympy())

class Plus(BinaryOperator):
    def toSympy(self):
        return super().toSympy('+')

class Divide(BinaryOperator):
    def toSympy(self):
        return super().toSympy('/')

class Times(BinaryOperator):
    def toSympy(self):
        return super().toSympy('*')

class Minus(BinaryOperator):
    def toSympy(self):
        return super().toSympy('-')

class Pow(BinaryOperator):
    def toSympy(self):
        return super().toSympy('**')

class Arrow(BinaryOperator):
    def toSympy(self):
        return '%s,%s' % (self.left.toSympy(), self.right.toSympy())

class Eq(BinaryOperator):
    def toSympy(self):
        return super().toSympy('-')



class UnaryOperator:
    def __init__(self, val):
        self.val = val

    def __str__(self):
        return '%s(%s)' % (self.__class__.__name__, self.val)

class Opp(UnaryOperator):
    def toSympy(self):
        return '('+ '-' + self.val.toSympy() +')'

class Fact(UnaryOperator):
    def toSympy(self):
        return '(' + self.val.toSympy() +'!)'

class Diff:
    def __init__(self, val, nb):
        self.val = val
        self.nb=nb

    def __str__(self):
        return 'Diff('+str(self.val)+','+str(self.nb)+')'

    def toSympy(self):
        return 'diff('+self.val.toSympy()+','+self.val.args[0].toSympy()+','+str(self.nb)+')'

class List:
    def __init__(self, l):
        self.list = l

    def __str__(self):
        if len(self.list)==0:
            return 'List([])'
        s = 'List(['+str(self.list[0])
        for e in self.list[1:]:
            s = s + ', ' + str(e)
        return s+'])'

    def __getitem__(self,index):
        return self.list[index]

    def __add__(self, other):
        return List(self.list+other.list)

    def __len__(self):
        return len(self.list)

    def getList(self):
        return self.list

    def toSympy(self):
        if len(self.list)==0:
            return ''
        s = self.list[0].toSympy()
        for e in self.list[1:]:
            s = s + ', ' + e.toSympy()
        return s

class FunctionCall:
    def __init__(self, function, args):
        self.function = function
        self.args = args

    def __str__(self):
        return 'FunctionCall('+str(self.function)+','+str(self.args)+')'

    def toSympy(self):
        if type(self.function)==Id:
            return self.translate(self.function.toSympy(), self.args)

    def translate(self, function, args):
        mathematicaToSympy={'Sqrt' : (lambda a: 'sqrt('+a[0].toSympy()+')'),
                            'Sin' : (lambda a: 'sin('+a[0].toSympy()+')'),
                            'Cos' : (lambda a: 'cos('+a[0].toSympy()+')'),
                            'Tan' : (lambda a: 'tan('+a[0].toSympy()+')'),
                            'Arccos' : (lambda a: 'acos('+a[0].toSympy()+')'),
                            'Arcsin' : (lambda a: 'asin('+a[0].toSympy()+')'),
                            'Arctan' : (lambda a: 'atan('+a[0].toSympy()+')'),
                            'Sum' : (lambda a: 'summation('+a[0].toSympy()+''.join(([(lambda l:',('+l[0].toSympy()+',('+ l[1].toSympy() +','+ l[2].toSympy() +'))')(l) for l in a[1:]]))+')'),
                            'Integrate' : (lambda a: 'integrate('+a[0].toSympy()+''.join(list(reversed([(lambda l:',('+l.toSympy()+')')(l) for l in a[1:]])))+')'),
                            'N' : (lambda a: 'N('+a.toSympy()+')'),
                            'D' : (lambda a: 'diff('+a[0].toSympy()+', '+', '.join([l.toSympy() for l in a[1:]])+')'),
                            'Exp' : (lambda a: 'exp('+a.toSympy()+')'),
                            'Simplify' : (lambda a: 'simplify('+a.toSympy()+')'),
                            'Power' : (lambda a: 'Pow('+a.toSympy()+')'),
                            'Log' : (lambda a: 'log('+List(list(reversed(a.getList()))).toSympy()+')'),
                            'Log10' : (lambda a: '(log('+a[0].toSympy()+'))/(log(10))'),
                            'Log2' : (lambda a: '(log('+a[0].toSympy()+'))/(log(2))'),
                            'Factorial' : (lambda a: '('+a[0].toSympy()+'!)'),
                            'Abs' : (lambda a: 'Abs('+a[0].toSympy()+')'),
                            'Ceiling' : (lambda a: 'ceiling('+a[0].toSympy()+')'),
                            'Floor' : (lambda a: 'floor('+a[0].toSympy()+')'),
                            'Limit' : (lambda a: 'limit('+a[0].toSympy() +','+ a[1].toSympy()+')'),
                            'Solve' : (lambda a: 'solve(['+a[0].toSympy() +'],['+ a[1].toSympy()+'])'),
                           }

        for name in mathematicaToSympy.keys():
            if name == function:
                return '('+mathematicaToSympy[name](args)+')'

        return '('+function+'('+ self.args.toSympy() +')'+')'

class Id:
    def __init__(self, id):
        self.id=id

    def __str__(self):
        return 'Id('+str(self.id)+')'

    def toSympy(self):
        return self.translateId(self.id)

    def translateId(self, id):
        mathematicaToSympy={'Infinity' : 'oo',
                            'I' : 'I',
                            'Pi' : 'pi',
                            'GoldenRatio' : 'GoldenRatio',
                            'EulerGamma' : 'EulerGamma',
                           }
        if id in mathematicaToSympy.keys():
            return mathematicaToSympy[id]
        return str(id)



# A lot of this code come from https://github.com/sympy/sympy_gamma/blob/master/app/logic/logic.py
# http://www.sympygamma.com/
import sys
import queue
import collections

import sympy
from sympy.core.function import FunctionClass
from sympy import latex, series, sympify, solve, Derivative, Integral, Symbol, diff, integrate
from sympy.parsing.sympy_parser import stringify_expr, eval_expr, standard_transformations, convert_xor, TokenError

from .parser import Parser
from .config import Config
from .supyprocess import process

PREEXEC = """from sympy import *
import sympy
from sympy.solvers.diophantine import diophantine
"""

def evaluate(s):
    result = None
    parser = Parser(s)
    inputFormula=parser.normalize()
    evaluated = process(eval_input, inputFormula, timeout=Config().timeout, heap_size=Config().max_heap)

    return evaluated
        
def eval_input(s):
    namespace = {}
    exec(PREEXEC, {}, namespace)

    def plot(f=None, **kwargs):
        pass
    namespace.update({
        'plot': plot,  # prevent textplot from printing stuff
        'help': lambda f: f
    })

    transformations = list(standard_transformations)
    parsed = stringify_expr(s, {}, namespace, transformations)
    try:
        evaluated = eval_expr(parsed, {}, namespace)
    except SyntaxError:
        raise
    except Exception as e:
        raise ValueError(str(e))

    return evaluated



# A lot of this code come from http://docs.sympy.org/dev/_modules/sympy/parsing/mathematica.html
# http://sympy.org/en/index.html

from .mathematicaYacc import mathematicaParser
from .mathematicaLex import mathematicaLexer
from .calchasPreprocessing import calchasToSympy
from re import match, sub

class Parser():
    def __init__(self, expr):
        self.expr=expr

    def fromCalchas(self):
        result = calchasToSympy('('+self.expr+')')
        if result:
            self.expr=result

    def fromMathematica(self):
        result = mathematicaParser.parse('('+self.expr+')', lexer=mathematicaLexer)
        if result:
            self.expr=result.toSympy()

    def normalize(self):
        self.fromMathematica()
        self.fromCalchas()
        return 'simplify('+self.expr+',2)'



"""Request handler of the module."""

from ppp_datamodel import Sentence, Resource
from ppp_datamodel.communication import TraceItem, Response, Request
from ppp_libmodule.exceptions import ClientError
from sympy import count_ops, latex
from sympy.parsing.sympy_parser import parse_expr

from .evaluator import evaluate
from .notation import relevance, isMath, traceContainsSpellChecker

class RequestHandler:
    def __init__(self, request):
        self.language = request.language
        self.measures = request.measures
        self.trace = request.trace
        self.tree = request.tree

    def answer(self):
        if not isinstance(self.tree, Sentence):
            return []

        if not isMath(self.tree.value) or traceContainsSpellChecker(self.trace):
            return []

        outputFormula=evaluate(self.tree.value)
        outputTree=Resource(latex(outputFormula), value_type='math-latex')
        measures = {
            'accuracy': 1,
            'relevance': relevance(self.tree.value, outputFormula)
        }
        trace = self.trace + [TraceItem('CAS', outputTree, measures)]
        response = Response(self.language, outputTree, measures, trace)
        return [response]



"""Processes handling helpers, imported from
https://github.com/ProgVal/Limnoria/
"""

import resource
import multiprocessing

processesSpawned = 1
class SupyProcess(multiprocessing.Process): # pragma: no cover
    def __init__(self, *args, **kwargs):
        global processesSpawned
        processesSpawned += 1
        super(SupyProcess, self).__init__(*args, **kwargs)

class CommandProcess(SupyProcess): # pragma: no cover
    """Just does some extra logging and error-recovery for commands that need
    to run in processes.
    """
    def __init__(self, target=None, args=(), kwargs={}):
        pn = kwargs.pop('pn', 'Unknown')
        cn = kwargs.pop('cn', 'unknown')
        procName = 'Process #%s (for %s.%s)' % (processesSpawned, pn, cn)
        self.__parent = super(CommandProcess, self)
        self.__parent.__init__(target=target, name=procName,
                               args=args, kwargs=kwargs)

    def run(self):
        self.__parent.run()
        
class ProcessTimeoutError(Exception): # pragma: no cover
    """Gets raised when a process is killed due to timeout."""
    pass
        
def process(f, *args, **kwargs): # pragma: no cover
    """Runs a function <f> in a subprocess.
    
    Several extra keyword arguments can be supplied. 
    <pn>, the pluginname, and <cn>, the command name, are strings used to
    create the process name, for identification purposes.
    <timeout>, if supplied, limits the length of execution of target 
    function to <timeout> seconds."""
    timeout = kwargs.pop('timeout', None)
    heap_size = kwargs.pop('heap_size', None)
    if resource and heap_size is None:
        heap_size = resource.RLIM_INFINITY
    
    try:
        q = multiprocessing.Queue()
    except OSError:
        log.error('Using multiprocessing.Queue raised an OSError.\n'
                'This is probably caused by your system denying semaphore\n'
                'usage. You should run these two commands:\n'
                '\tsudo rmdir /dev/shm\n'
                '\tsudo ln -Tsf /{run,dev}/shm\n'
                '(See https://github.com/travis-ci/travis-core/issues/187\n'
                'for more information about this bug.)\n')
        raise
    def newf(f, q, *args, **kwargs):
        if resource:
            rsrc = resource.RLIMIT_DATA
            resource.setrlimit(rsrc, (heap_size, heap_size))
        try:
            r = f(*args, **kwargs)
            q.put(r)
        except Exception as e:
            q.put(e)
    targetArgs = (f, q,) + args
    p = CommandProcess(target=newf, args=targetArgs, kwargs=kwargs)
    
    p.start()
    p.join(timeout)
    if p.is_alive():
        p.terminate()
        q.close()
        raise ProcessTimeoutError("%s aborted due to timeout." % (p.name,))
    try:
        v = q.get(block=False)
    except queue.Empty:
        return None
    finally:
        q.close()
    if isinstance(v, Exception):
        raise v
    else:
        return v



import ply.yacc as yacc
from .calchasLex import tokens
from .calchasTree import Plus, Minus, Times, Opp, FunctionCall, List, Divide, Pow, Id, Fact, Mod, Eq
from sympy import latex

precedence = (
    ('nonassoc', 'EQ'),
    ('nonassoc', 'NUMBER'),
    ('nonassoc', 'ID'),
    ('left', 'PLUS', 'MINUS'),
    ('left', 'TIMES', 'DIVIDE'),
    ('left', 'UMINUS'),
    ('left', 'UPLUS'),
    ('left', 'MOD'),
    ('right', 'POW'),
    ('right', 'DTIMES'),
    ('left', 'EXCL'),
    ('nonassoc', 'RPAREN'),
    ('nonassoc', 'LPAREN'),
    ('nonassoc', 'COMMA'),
)

start = 'expression'

def p_expr_unary(p):
    '''expression : MINUS expression %prec UMINUS
                  | PLUS expression %prec UPLUS'''
    if p[1] == '-':
        p[0] = Opp(p[2])
    else:
        p[0] = p[2]

def p_expression_arith(p):
    '''expression : expression PLUS expression
                  | expression MINUS expression
                  | expression TIMES expression
                  | expression POW expression
                  | expression MOD expression
                  | expression DIVIDE expression
                  | expression DTIMES expression'''
    tokenToNode = { '+' : Plus,
                    '-' : Minus,
                    '^' : Pow,
                    '/' : Divide,
                    '*' : Times,
                    '%' : Mod,
                    '**' : Pow,
                  }
    p[0] = tokenToNode[p[2]](p[1], p[3])

def p_expression_fact(p):
    '''expression : expression EXCL'''
    p[0] = Fact(p[1])

def p_expression_parentheses(p):
    '''expression : LPAREN expression RPAREN'''
    p[0] = p[2]

def p_expression_eq(p):
    '''expression : expression EQ expression
                  | expression EQ EQ expression'''
    if p[3] == '=':
        p[0] = Eq(p[1], p[4])
    else:
        p[0] = Eq(p[1], p[3])

def p_function_call_arg(p):
    '''expression : ID LPAREN expression_list RPAREN'''
    p[0] = FunctionCall(Id(p[1]), p[3])

def p_function_call_empty(p):
    '''expression : ID LPAREN RPAREN'''
    p[0] = FunctionCall(Id(p[1]), List([]))

def p_expression_list(p):
    '''expression_list : expression'''
    p[0] = List([p[1]])

def p_expression_list_rec(p):
    '''expression_list : expression_list COMMA expression'''
    p[0] = p[1] + List([p[3]])

def p_expression_num(p):
    '''expression : NUMBER'''
    p[0] = Id(p[1])
    
def p_expression_id(p):
    '''expression : ID'''
    p[0] = Id(p[1])

def p_error(p):
    pass

calchasParser = yacc.yacc(debug=0, write_tables=0)



"""Configuration module."""
import os
import json
import logging
from ppp_libmodule.config import Config as BaseConfig
from ppp_libmodule.exceptions import InvalidConfig

class Config(BaseConfig):
    __slots__ = ('max_heap', 'timeout')
    config_path_variable = 'PPP_CAS_CONFIG'
    
    def parse_config(self, data):
        self.max_heap = data['max_heap']
        self.timeout = data['timeout']



from ppp_cas.calchasLex import calchasLexer
from ppp_cas.calchasYacc import calchasParser
    
def preprocessImplicitMultiplication(formula):
    calchasLexer.input(formula)
    previous = calchasLexer.token()
    output = previous.value
    while True:
        tok = calchasLexer.token()
        if not tok:
            break
        if (previous.type == 'RPAREN' and tok.type == 'LPAREN') or(previous.type == 'RPAREN' and tok.type == 'ID') or (previous.type == 'NUMBER' and tok.type == 'LPAREN') or (previous.type in ['ID', 'NUMBER'] and tok.type in ['ID', 'NUMBER']):
            output = output + '*'
        output = output + tok.value
        previous = tok
    return output
    
def parseCalchas(formula):
    return calchasParser.parse(preprocessImplicitMultiplication(formula), lexer=calchasLexer)

def calchasToSympy(formula):
    val = parseCalchas(formula)
    if val == None:
        return None
    else:
        return val.toSympy()



import re

class BinaryOperator:
    def __init__(self, left, right):
        self.left = left
        self.right = right

    def __str__(self):
        return '%s(%s, %s)' % (self.__class__.__name__, self.left, self.right)

    def toSympy(self, op):
        return '(%s%s%s)' % (self.left.toSympy(), op, self.right.toSympy())

class Plus(BinaryOperator):
    def toSympy(self):
        return super().toSympy('+')

class Divide(BinaryOperator):
    def toSympy(self):
        return super().toSympy('/')

class Mod(BinaryOperator):
    def toSympy(self):
        return super().toSympy('%')

class Times(BinaryOperator):
    def toSympy(self):
        return super().toSympy('*')

class Minus(BinaryOperator):
    def toSympy(self):
        return super().toSympy('-')

class Pow(BinaryOperator):
    def toSympy(self):
        return super().toSympy('**')



class UnaryOperator:
    def __init__(self, val):
        self.val = val

    def __str__(self):
        return '%s(%s)' % (self.__class__.__name__, self.val)

class Opp(UnaryOperator):
    def toSympy(self):
        return '('+ '-' + self.val.toSympy() +')'

class Fact(UnaryOperator):
    def toSympy(self):
        return '(gamma(' + self.val.toSympy() +'+1))'

class List:
    def __init__(self, l):
        self.list = l

    def __str__(self):
        if len(self.list)==0:
            return 'List([])'
        s = 'List(['+str(self.list[0])
        for e in self.list[1:]:
            s = s + ', ' + str(e)
        return s+'])'

    def __getitem__(self,index):
        return self.list[index]

    def __add__(self, other):
        return List(self.list+other.list)

    def __len__(self):
        return len(self.list)

    def getList(self):
        return self.list

    def toSympy(self):
        if len(self.list)==0:
            return ''
        s = self.list[0].toSympy()
        for e in self.list[1:]:
            s = s + ', ' + e.toSympy()
        return s

    def toSympyRightAssociative(self):
        def rightParentheses(l):
            if len(l) == 0:
                return ''
            if len(l) == 1:
                return l[0].toSympy()
            return l[0].toSympy()+',('+rightParentheses(l[1:])+')'
        return rightParentheses(self.list)

class Eq(BinaryOperator):
    def toSympy(self):
        return super().toSympy('-')

class FunctionCall:
    def __init__(self, function, args):
        self.function = function
        self.args = args

    def __str__(self):
        return 'FunctionCall('+str(self.function)+','+str(self.args)+')'

    def toSympy(self):
        if type(self.function) == Id:
            return self.translate(self.function.toSympy(), self.args)

    def translate(self, function, args):
        mathematicaToSympy={r'^[aA]bs$' : (lambda a: 'Abs('+a.toSympy()+')'),
                            r'^[mM]od$' : (lambda a: 'Abs('+a.toSympy()+')'),
                            r'^[sS]ig(n)?$' : (lambda a: 'sign('+a.toSympy()+')'),
                            r'^[sS]gn$' : (lambda a: 'sign('+a.toSympy()+')'),
                            r'^[sS]ignum$' : (lambda a: 'sign('+a.toSympy()+')'),
                            r'^[pP]ow(er)?$' : (lambda a: 'Pow('+a.toSympy()+')'),
                            r'^[sS]qrt$' : (lambda a: 'sqrt('+a.toSympy()+')'),
                            r'^[rR]oot$' : (lambda a: 'root('+a.toSympy()+')'),
                            r'^[eE]xp$' : (lambda a: 'exp('+a.toSympy()+')'),
                            r'^[lL](n|og)$' : (lambda a: 'log('+a.toSympy()+')'),
                            r'^[lL]og10$' : (lambda a: 'log('+a.toSympy()+',10)'),
                            r'^[lL]og2$' : (lambda a: 'log('+a.toSympy()+',2)'),
                            r'^[lL]g$' : (lambda a: 'log('+a.toSympy()+',10)'),
                            r'^[lL]b$' : (lambda a: 'log('+a.toSympy()+',2)'),
                            r'^(Gamma|GAMMA)$' : (lambda a: 'gamma('+a.toSympy()+')'),
                            r'^[fF]act(orial)?$' : (lambda a: 'gamma('+a.toSympy()+'+1)'),
                            r'^[cC]os$' : (lambda a: 'cos('+a.toSympy()+')'),
                            r'^[sS]in$' : (lambda a: 'sin('+a.toSympy()+')'),
                            r'^[tT]an$' : (lambda a: 'tan('+a.toSympy()+')'),
                            r'^[cC](osec|sc)$' : (lambda a: 'csc('+a.toSympy()+')'),
                            r'^[sS]ec$' : (lambda a: 'sec('+a.toSympy()+')'),
                            r'^[cC]ot(an)?$' : (lambda a: 'cot('+a.toSympy()+')'),
                            r'^[aA](rc)?[cC]os$' : (lambda a: 'acos('+a.toSympy()+')'),
                            r'^[aA](rc)?[sS]in$' : (lambda a: 'asin('+a.toSympy()+')'),
                            r'^[aA](rc)?[tT]an$' : (lambda a: 'atan('+a.toSympy()+')'),
                            r'^[aA](rc)?[cC](sc|osec)$' : (lambda a: 'asin(1/('+a.toSympy()+'))'),
                            r'^[aA](rc)?[sS]ec$' : (lambda a: 'acos(1/('+a.toSympy()+'))'),
                            r'^[aA](rc)?[cC]ot(an)?$' : (lambda a: 'acot('+a.toSympy()+')'),
                            r'^[cC](os)?h$' : (lambda a: 'cosh('+a.toSympy()+')'),
                            r'^[sS](in)?h$' : (lambda a: 'sinh('+a.toSympy()+')'),
                            r'^[tT](an)?h$' : (lambda a: 'tanh('+a.toSympy()+')'),
                            r'^[cC](osec|sc)h$' : (lambda a: '1/(sinh('+a.toSympy()+')'),
                            r'^[sS]ech$' : (lambda a: '1/(cosh('+a.toSympy()+')'),
                            r'^[cC]ot(an)?h$' : (lambda a: 'coth('+a.toSympy()+')'),
                            r'^[aA](r[cg]?)?[cC](os)?h$' : (lambda a: 'acosh('+a.toSympy()+')'),
                            r'^[aA](r[cg]?)?[sS](in)?h$' : (lambda a: 'asinh('+a.toSympy()+')'),
                            r'^[aA](r[cg]?)?[tT](an)?h$' : (lambda a: 'atan('+a.toSympy()+')'),
                            r'^[aA](r[cg]?)?[cC](sc|osec)h$' : (lambda a: 'asinh(1/('+a.toSympy()+'))'),
                            r'^[aA](r[cg]?)?[sS]ech$' : (lambda a: 'acosh(1/('+a.toSympy()+'))'),
                            r'^[aA](r[cg]?)?[cC]ot(an)?h$' : (lambda a: 'acoth('+a.toSympy()+')'),
                            r'^[fF]loor$' : (lambda a: 'floor('+a.toSympy()+')'),
                            r'^[cC]eil(ing)?$' : (lambda a: 'ceiling('+a.toSympy()+')'),
                            r'^[dD]igamma$' : (lambda a: 'digamma('+a.toSympy()+')'),
                            r'^[bB]eta$' : (lambda a: 'beta('+a.toSympy()+')'),
                            r'^[bB]inomial$' : (lambda a: 'gamma('+a[0].toSympy()+'+1)/(gamma('+a[1].toSympy()+'+1)*gamma('+a[0].toSympy()+'-'+a[1].toSympy()+'+1))'),
                            r'^C$' : (lambda a: 'gamma('+a[0].toSympy()+'+1)/(gamma('+a[1].toSympy()+'+1)*gamma('+a[0].toSympy()+'-'+a[1].toSympy()+'+1))'),
                            r'^[cC]omb(ination)?$' : (lambda a: 'gamma('+a[0].toSympy()+'+1)/(gamma('+a[1].toSympy()+'+1)*gamma('+a[0].toSympy()+'-'+a[1].toSympy()+'+1))'),
                            r'^A$' : (lambda a: 'gamma('+a[1].toSympy()+'+1)*(gamma('+a[0].toSympy()+'+1)/(gamma('+a[1].toSympy()+'+1)*gamma('+a[0].toSympy()+'-'+a[1].toSympy()+'+1)))'),
                            r'^[pP]artial[pP]ermutation$' : (lambda a: 'gamma('+a[1].toSympy()+'+1)*(gamma('+a[0].toSympy()+'+1)/(gamma('+a[1].toSympy()+'+1)*gamma('+a[0].toSympy()+'-'+a[1].toSympy()+'+1)))'),
                            r'^[gG]c[dm]$' : (lambda a: 'gcd('+a.toSympy()+')'),
                            r'^[hH]cf$' : (lambda a: 'gcd('+a.toSympy()+')'),
                            r'^[lL]c[mM]$' : (lambda a: 'lcm('+a.toSympy()+')'),
                            r'^[dD](iff|eriv(e|at(e|ive)))?$' : (lambda a: 'diff('+a.toSympy()+')'),
                            r'^[iI]nt(egra(te|l))?$' : (lambda a: 'integrate('+a.toSympyRightAssociative()+')'),
                            r'^[aA]ntiderivative$' : (lambda a: 'integrate('+a.toSympyRightAssociative()+')'),
                            r'^[sS]um(mation)?' : (lambda a: 'summation('+a.toSympyRightAssociative()+')'),
                            r'^[aA]pprox(imation)?$' : (lambda a: 'N('+a.toSympy()+')'),
                            r'^N$' : (lambda a: 'N('+a.toSympy()+')'),
                            r'^[nN]umeric$' : (lambda a: 'N('+a.toSympy()+')'),
                            r'^[eE]val(f)?$' : (lambda a: 'N('+a.toSympy()+')'),
                            r'^[sS]impl(if(y|ication))?$' : (lambda a: 'simplify('+a.toSympy()+')'),
                            r'^[sS]ol(ve|ution(s)?)?$' : (lambda a: 'solve(['+a[0].toSympy() +'],['+ a[1].toSympy()+'])'),
                            r'^[lL]im(it)?$' : (lambda a: 'limit('+a.toSympy()+')'),
                            r'^[lL]im(it)?[lL]$' : (lambda a: 'limit('+a.toSympy()+',dir=\'-\')'),
                            r'^[lL]im(it)?[rR]$' : (lambda a: 'limit('+a.toSympy()+',dir=\'+\')'),
                           }

        for pattern in mathematicaToSympy.keys():
            if re.match(pattern, function):
                return '('+mathematicaToSympy[pattern](args)+')'
        return '('+function+'('+ self.args.toSympy() +')'+')'

class Id:
    def __init__(self, id):
        self.id=id

    def __str__(self):
        return 'Id('+str(self.id)+')'

    def toSympy(self):
        return self.translateId(self.id)

    def translateId(self, name):
        if type(name) == int:
            return str(name)
        mathematicaToSympy={r'^[iI]nfinity$' : 'oo',
                            r'^[iI]nfty$' : 'oo',
                            r'^I$' : 'I',
                            r'^[pP]i$' : 'pi',
                            r'^(GoldenRatio|phi|Phi)$' : 'GoldenRatio',
                            r'^([eE]uler[gG]amma|gamma)$' : 'EulerGamma',
                            r'^[eE]$' : 'E',
                           }
        for pattern in mathematicaToSympy.keys():
            if re.match(pattern, name):
                return mathematicaToSympy[pattern]
        return str(name)



import ply.lex as lex

tokens = (
    'NUMBER',
    'PLUS',
    'MINUS',
    'TIMES',
    'DIVIDE',
    'LPAREN',
    'RPAREN',
    'COMMA',
    'ID',
    'EXCL',
    'POW',
    'DTIMES',
    'MOD',
    'EQ',
)

t_DTIMES     = r'\*\*'
t_PLUS       = r'\+'
t_MINUS      = r'-'
t_TIMES      = r'\*'
t_DIVIDE     = r'/'
t_LPAREN     = r'\('
t_RPAREN     = r'\)'
t_COMMA      = r','
t_EXCL       = r'!'
t_POW        = r'\^'
t_MOD        = r'%'
t_EQ         = r'='

def t_NUMBER(t):
    r'([0-9]*\.[0-9]+|[0-9]+)'
    return t
    
def t_ID(t):
    r'[a-zA-Z][a-zA-Z0-9]*'
    return t

def t_newline(t):
    r'\n+'
    t.lexer.lineno += len(t.value)

t_ignore  = ' \t'

def t_error(t):
    t.lexer.skip(1)

calchasLexer = lex.lex()




CARACTERISTIC_SYMBOLS = ['sqrt'] + list(r'\/*-()_^+[]{}=!%0123456789')
def isMath(formula):
    return any(e in formula for e in CARACTERISTIC_SYMBOLS)

def traceContainsSpellChecker(trace):
    return any(e['module'] == 'spell-checker' for e in trace)

def relevance(inputFormula, outputTree):
    lenOutput = len(str(outputTree))
    if lenOutput != 0:
        return len(inputFormula)/lenOutput
    return 0
